from __future__ import annotations

import ast
import inspect
import os
import shutil
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, cast

from django.db.models import Model

import isort
from black import FileMode, format_str

from ..exceptions import ConversionError
from .objects import AppModel, AppView
from .utils import collect_references, ensure_http_response, import_from_path


if TYPE_CHECKING:
    from pathlib import Path

    from ..app import Django


class Resolver:
    imports: set[str]
    local_refs: set[str]
    global_refs: set[str]

    def __init__(self, converter: Converter, module_name: str):
        self.converter = converter
        self.module_name = module_name

        self.imports = set()
        self.local_refs = set()
        self.global_refs = set()

    def add(self, name: str, references: set[str]):
        # Log this object definition
        self.converter.imports[name] = f"from {self.module_name} import {name}"
        self.local_refs.add(name)

        self.add_references(references)

    def add_references(self, references: set[str]):
        # Collect import and global references
        for ref in references:
            if ref in self.local_refs:
                pass
            elif ref in self.converter.imports:
                self.imports.add(self.converter.imports[ref])
            else:
                self.global_refs.add(ref)

    def gen_src(self):
        all_src = []
        global_refs = self.global_refs.copy()
        while self.global_refs:
            # Collect the source for the reference, and grab any references it may have
            global_ref = self.global_refs.pop()
            src, references = self.converter.collect_definition(global_ref)
            all_src.append(src)

            # Record references from this
            self.add_references(references)

            # We may have added references to locals we have seen before - remove them
            self.global_refs = self.global_refs - self.local_refs

            # The global reference is now local
            self.local_refs.add(global_ref)

            # Tell other modules where to find it
            self.converter.imports[global_ref] = (
                f"from {self.module_name} import {global_ref}"
            )

        # Restore global refs
        self.global_refs = global_refs
        return "\n".join(list(self.imports) + all_src)


class Converter:
    #: AppModel instances for models that are being moved to models.py
    models: list[AppModel]

    #: AppView instances for views that are being moved to views.py
    views: list[AppView]

    #: Import paths - existing imports in the app, plus converted models and views
    imports: dict[str, str]

    #: Definitions in the module's top level scope which have been converted
    used: set[str]

    def __init__(self, app: Django, path: Path, name: str):
        self.app = app
        self.root_path = path
        self.project_name = name
        self.module = app.app_module
        self.src = inspect.getsource(app.app_module)
        self.ast = ast.parse(self.src)
        self.used = set()

        self.collect_imports()

    def write_file(self, filename: str | Path, *content):
        path = Path(filename)
        formatted = isort.code(format_str("\n".join(content), mode=FileMode()))
        path.write_text(formatted)

    def write(self) -> None:
        self.build_project()
        self.build_app_models()
        self.build_app_views()
        self.build_app_urls()
        self.build_app_admin()
        self.build_app_unused()

    def build_project(self) -> None:
        # Copy the env, so we're still working within any venv
        env = os.environ.copy()
        env.pop("DJANGO_SETTINGS_MODULE", None)
        try:
            result = subprocess.run(
                ["django-admin", "startproject", self.project_name, self.root_path],
                env=env,
                capture_output=True,
                text=True,
            )
        except Exception as e:
            raise ConversionError(f"Could not run django-admin: {e}")

        if result.returncode != 0:
            raise ConversionError(f"django-admin startproject failed: {result.stderr}")

        # Create app dir
        app_dir = self.root_path / self.project_name / self.app.app_name
        app_dir.mkdir()
        (app_dir / "__init__.py").touch()

        self.build_settings()
        self.build_urls()
        self.copy_assets()

    def build_settings(self) -> None:
        """
        Collect settings from the app definition and update the project settings
        """
        resolver = Resolver(self, f"{self.project_name}.settings")

        # Collect from app definition and remove nanodjango-specific settings
        app_settings = {}
        for node in self.ast.body:
            # Look for app = Django(..)
            if (
                isinstance(node, ast.Assign)
                and any(
                    [
                        target.id == "app"
                        for target in node.targets
                        if isinstance(target, ast.Name)
                    ]
                )
                and isinstance(node.value, ast.Call)
                and isinstance(node.value.func, ast.Name)
                and node.value.func.id == "Django"
            ):
                for keyword in node.value.keywords:
                    name: str = cast(str, keyword.arg)
                    if name.isupper():
                        # Exclude nanodjango-specific settings
                        if name in [
                            "ADMIN_URL",
                            "EXTRA_APPS",
                            "SQLITE_DATABASE",
                            "MIGRATIONS_DIR",
                        ]:
                            continue
                        else:
                            app_settings[name] = keyword.value
                            resolver.add_references(collect_references(keyword.value))

        # Load settings file
        filename = self.root_path / self.project_name / "settings.py"
        settings = import_from_path("nanodjango.convert.tmp_settings", filename)
        settings_src = inspect.getsource(settings)
        settings_ast = ast.parse(settings_src)

        # Replace settings
        for node in settings_ast.body:
            if isinstance(node, ast.Assign):
                for target in node.targets:
                    if not isinstance(target, ast.Name):
                        continue
                    if target.id in app_settings:
                        node.value = app_settings.pop(target.id)

                    # Special case for INSTALLED_APPS
                    if target.id == "INSTALLED_APPS" and isinstance(
                        node.value, ast.List
                    ):
                        # Inject our converted app
                        node.value.elts.append(
                            ast.Constant(
                                value=f"{self.project_name}.{self.app.app_name}"
                            )
                        )

                        # Append any EXTRA_APPS
                        if "EXTRA_APPS" in self.app._settings:
                            for extra_app in self.app._settings["EXTRA_APPS"]:
                                node.value.elts.append(ast.Constant(value=extra_app))

        # Add any extra vars
        for name, value in app_settings.items():
            node = ast.Assign(targets=[ast.Name(id=name, ctx=ast.Store())], value=value)
            settings_ast.body.append(node)

        # Insert any references - usually imports
        ref_src = resolver.gen_src()
        if ref_src:
            imports = set()
            others = set()

            for node in ast.parse(ref_src).body:
                if isinstance(node, (ast.Import, ast.ImportFrom)):
                    imports.add(node)
                else:
                    others.add(node)

            # We're expecting a comment at the top - insert imports just after that
            if (
                len(settings_ast.body) == 0
                or not isinstance(settings_ast.body[0], ast.Expr)
                or not isinstance(settings_ast.body[0].value, ast.Constant)
            ):
                raise ConversionError("Unexpected start to settings.py")
            settings_ast.body[1:1] = imports

            # Insert other definitions after the last import
            index = 0
            for node in settings_ast.body:
                if isinstance(node, (ast.Expr, ast.Import, ast.ImportFrom)):
                    index += 1
                    continue
                break
            settings_ast.body[index:index] = others

        # Save settings
        self.write_file(filename, ast.unparse(settings_ast))

    def build_urls(self) -> None:
        filename = self.root_path / self.project_name / "urls.py"
        src = filename.read_text()

        # Add path to app
        pattern = "urlpatterns = ["
        if pattern not in src:
            raise ConversionError("Expected to find urlpatterns in urls.py")
        src = src.replace(
            "from django.urls import path",
            "from django.urls import include, path",
        ).replace(
            pattern,
            f'{pattern}\n    path("", include("{self.project_name}.{self.app.app_name}.urls")),',
        )

        if "ADMIN_URL" in self.app._settings:
            pattern = '"admin/"'
            if pattern not in src:
                raise ConversionError("Expected to find admin path in urls.py")
            src = src.replace(pattern, f'"{self.app._settings["ADMIN_URL"]}"')

        self.write_file(filename, src)

    def copy_assets(self) -> None:
        """
        Copy static, templates, and migrations into app, and db.sqlite3 if it exists
        """
        script_dir = self.app.app_path.parent

        db_file = script_dir / self.app._settings.get("SQLITE_DATABASE", "db.sqlite3")
        if db_file.exists():
            shutil.copy(db_file, self.root_path / "db.sqlite3")

        dir_names = [
            "static",
            "templates",
            self.app._settings.get("MIGRATIONS_DIR", "migrations"),
        ]
        for dir_name in dir_names:
            src_dir = script_dir / dir_name
            if src_dir.exists() and src_dir.is_dir():
                shutil.copytree(
                    src_dir,
                    self.root_path / self.project_name / self.app.app_name / dir_name,
                )

    def collect_imports(self) -> dict[str, str]:
        """
        Collect a lookup for imported names

        Values are the import strings - we'll use isort to merge modules later
        """
        self.imports = {}
        for node in self.ast.body:
            if isinstance(node, ast.Import):
                for alias in node.names:
                    self.imports[alias.name] = f"import {alias.name}"
            elif isinstance(node, ast.ImportFrom):
                for alias in node.names:
                    self.imports[alias.name] = f"from {node.module} import {alias.name}"

        return self.imports

    def collect_definition(self, obj_name) -> tuple[str, set[str]]:
        """
        Collect a definition of a module top-level scope object
        """
        if obj_name == "ensure_http_response":
            # Inject its dependencies
            self.imports.update(getattr(ensure_http_response, "_dependencies", {}))
            obj_src = inspect.getsource(ensure_http_response)
            obj_ast = ast.parse(obj_src)
            references = collect_references(obj_ast.body[0])
            return obj_src, references

        if obj_name not in self.module.__dict__:
            raise ConversionError(f"Reference to unknown symbol {obj_name}")
        obj = self.module.__dict__[obj_name]

        # Try to build src from the object
        try:
            obj_src = inspect.getsource(obj)
        except TypeError:
            # Not a class, method, function, or code object
            pass
        else:
            self.used.add(obj_name)
            obj_ast = ast.parse(obj_src)
            references = collect_references(obj_ast.body[0])
            return obj_src, references

        # Look for an assignment
        for node in self.ast.body:
            if isinstance(node, ast.Assign):
                for target in node.targets:
                    if isinstance(target, ast.Name) and target.id == obj_name:
                        obj_src = ast.unparse(node).strip()
                        self.used.add(obj_name)
                        obj_ast = ast.parse(obj_src)
                        references = collect_references(obj_ast.body[0])
                        return obj_src, references

        # TODO: We could have more exhaustive definition collection
        raise ConversionError(f"Reference to undetermined symbol {obj_name}")

    def build_app_models(self) -> None:
        self.models = []
        resolver = Resolver(self, ".models")

        for name, obj in self.module.__dict__.items():
            if inspect.isclass(obj) and issubclass(obj, Model):
                app_model = AppModel(name, obj)
                self.models.append(app_model)
                resolver.add(name, app_model.references)

        if not self.models:
            return

        self.write_file(
            self.root_path / self.project_name / self.app.app_name / "models.py",
            resolver.gen_src(),
            "\n".join([app_model.src for app_model in self.models]),
        )

    def build_app_views(self) -> None:
        self.views = []
        resolver = Resolver(self, ".views")

        for pattern, view in self.app._routes.items():
            app_view = AppView(pattern, view)
            self.views.append(app_view)
            resolver.add(view.__name__, app_view.references)

        if not self.views:
            return

        self.write_file(
            self.root_path / self.project_name / self.app.app_name / "views.py",
            resolver.gen_src(),
            "\n".join([app_view.src for app_view in self.views]),
        )

    def build_app_urls(self) -> None:
        urls = []
        for app_view in self.views:
            urls.append(app_view.make_url())

        if not urls:
            return

        self.write_file(
            self.root_path / self.project_name / self.app.app_name / "urls.py",
            "from django.urls import path",
            "from . import views",
            "urlpatterns = [",
            "\n".join([src for src in urls]),
            "]",
        )

    def build_app_admin(self) -> None:
        model_names: list[str] = []
        admin_srcs: list[str] = []
        resolver = Resolver(self, ".admin")

        # Collect @app.admin decorated models
        for app_model in self.models:
            if not app_model.admin_decorator:
                continue
            model_names.append(app_model.name)
            admin_srcs.append(app_model.make_model_admin())
            resolver.add_references(set([app_model.name]))

        # TODO: We could collect ModelAdmin. For now lets let it fall into unused

        if not model_names:
            return

        self.write_file(
            self.root_path / self.project_name / self.app.app_name / "admin.py",
            "from django.contrib import admin",
            resolver.gen_src(),
            "\n".join([src for src in admin_srcs]),
        )

    def build_app_unused(self) -> None:
        resolver = Resolver(self, ".unused")

        # We're not going to worry about unused imports
        self.used.update(self.imports.keys())

        # Also going to ignore the app
        self.used.add("app")

        unused = set(self.module.__dict__.keys()) - self.used
        all_src = []
        for obj_name in unused:
            if obj_name.startswith("_"):
                continue

            # Collect the source and objects it references
            obj_src, references = self.collect_definition(obj_name)
            all_src.append(obj_src)
            resolver.add(obj_name, references)

        if not all_src:
            return

        self.write_file(
            self.root_path / self.project_name / self.app.app_name / "unused.py",
            "# Definitions that were not used by the nanodjango converter",
            "# These will need to be merged into the rest of the app manually",
            resolver.gen_src(),
            "\n".join([src for src in all_src]),
        )

        print(f"Unused code detected, see {self.app.app_name}/unused.py")



from __future__ import annotations

import inspect
import os
import sys
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable

from django import setup
from django.contrib import admin
from django.urls import path as url_path
from django.views import View

from . import app_meta
from .exceptions import ConfigurationError, UsageError
from .urls import urlpatterns
from .views import string_view


if TYPE_CHECKING:
    from pathlib import Path

    from django.db.models import Model


class Django:
    """
    The main Django app
    """

    #: Name of the app script
    app_name: str

    #: Reference to the app script's module
    app_module: ModuleType

    #: Path of app script
    app_path: Path

    # Caches to aid ``convert``
    _settings: dict[str, Any] = {}
    _routes: dict[str, Callable] = {}

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)

        # Set app meta
        app_name = inspect.stack()[1].frame.f_globals["__name__"]
        app_meta._app_module = sys.modules[app_name]
        return instance

    def __init__(self, **_settings):
        """
        Initialise a new Django app, optionally with settings

        Usage::

            app = Django()
            app = Django(SECRET_KEY="some-secret", ALLOWED_HOSTS=["my.example.com"])
        """
        self._config(_settings)

    def _config(self, _settings):
        """
        Configure settings and patch Django ready for model definitions
        """
        self._settings = app_meta._app_conf = _settings

        # Settings
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "nanodjango.settings")
        from django.conf import settings

        self.settings = settings

        for key, value in _settings.items():
            setattr(settings, key, value)

        self.app_name = settings.DF_APP_NAME
        self.app_module = app_meta.get_app_module()
        self.app_path = Path(inspect.getfile(self.app_module))

        # Import and apply glue after django.conf has its settings
        from .django_glue.apps import prepare_apps
        from .django_glue.db import patch_db

        patch_db(self.app_name)
        prepare_apps(self.app_name, self.app_module)

        # Ready for Django's standard setup
        setup()

    def _prepare(self):
        """
        Perform any final setup for this project after it has been imported:

        * register the admin site
        """
        admin_url = self.settings.ADMIN_URL
        if admin_url:
            if not isinstance(admin_url, str) or not admin_url.endswith("/"):
                raise ConfigurationError(
                    "settings.ADMIN_URL must be a string path ending in /"
                )
            urlpatterns.append(url_path(admin_url.removeprefix("/"), admin.site.urls))

    def route(self, pattern: str):
        """
        Decorator to add a view to the urls

        The pattern should use the Django path syntax - see
        https://docs.djangoproject.com/en/dev/ref/urls/#path

        Usage::

            @app.route("/")
            def view(request):
                return "Hello"

        All paths are relative to the root URL, leading slashes will be ignored.
        """
        # We want to support Flask-like patterns which have leading / in its patterns.
        # Django does not use these, so strip them out.
        if pattern.startswith("/"):
            pattern = pattern[1:]

        def wrapped(fn):
            # Store route for convert lookup
            self._routes[pattern] = fn

            # Prepare CBVs
            if inspect.isclass(fn) and issubclass(fn, View):
                fn = fn.as_view()

            # Register URL
            urlpatterns.append(
                url_path(pattern.removeprefix("/"), string_view(fn), name=fn.__name__)
            )
            return fn

        return wrapped

    @property
    def has_admin(self):
        return isinstance(self.settings.ADMIN_URL, str)

    def admin(self, model: type[Model] | None = None, **options):
        """
        Decorator to add a model to the admin site

        The admin site must be added using ``settings.ADMIN_URL``.
        """
        if not self.has_admin:
            raise ConfigurationError(
                "Cannot register ModelAdmin - settings.ADMIN_URL is not set"
            )

        def wrap(model: type[Model]):
            admin.site.register(model, **options)
            return model

        if model is None:
            # Called with arguments, @admin(attr=val)
            return wrap

        # Called without arguments, @admin - call wrapped immediately
        return wrap(model)

    def run(self, args: list[str] | None = None):
        """
        Run a Django management command, passing all arguments

        Defaults to:
            runserver 0:8000
        """
        # Check if this is being called from click commands or directly
        if self.app_name not in sys.modules:
            # Hasn't been run through the ``nanodjango`` command
            if (
                "__main__" not in sys.modules
                or getattr(sys.modules["__main__"], "app") != self
            ):
                # Doesn't look like it was run directly either
                raise UsageError("App module not initialised")

            # Run directly, so register app module so Django won't try to load it again
            sys.modules[self.app_name] = sys.modules["__main__"]

        # Be helpful and check sys.argv for args. This will almost certainly be because
        # it's running directly.
        if args is None:
            args = sys.argv[1:]

        self._prepare()
        from django.core.management import execute_from_command_line

        if not args:
            args = ["runserver", "0:8000"]
        args = ["nanodjango"] + list(args)
        execute_from_command_line(args)

    def convert(self, path: Path, name: str):
        from .convert import Converter

        if path.exists():
            raise UsageError("Upgrade path is not empty - path cannot exist")
        path.mkdir()

        converter = Converter(app=self, path=path, name=name)
        converter.write()

    def __call__(self, *args, **kwargs):
        from django.core.wsgi import get_wsgi_application

        if "DEBUG" not in self._settings:
            from django.conf import settings

            settings.DEBUG = False

        application = get_wsgi_application()
        return application(*args, **kwargs)



import shutil
import sys
from importlib.machinery import SourceFileLoader
from pathlib import Path

import click


def load_app(path: str | Path):
    app = Path(path).absolute()
    sys.path.append(str(app.parent))
    module = SourceFileLoader(app.stem, str(app)).load_module()
    return getattr(module, "app")


@click.group()
@click.argument("app", type=Path, required=True)
@click.pass_context
def cli(ctx, app):
    # Import the script
    app = load_app(app)
    ctx.obj["app"] = app


@cli.command()
@click.argument("args", type=str, required=False, nargs=-1)
@click.pass_context
def run(ctx, args: tuple[str]):
    ctx.obj["app"].run(args)


@cli.command()
@click.argument("path", type=click.Path(), required=True)
@click.option("--name", default="project", help="The project name")
@click.option(
    "--delete",
    "can_delete",
    is_flag=True,
    help="If the target path is not empty, delete it before proceeding",
)
@click.pass_context
def convert(ctx, path: click.Path, name: str, can_delete: bool = False):
    target_path: Path = Path(str(path)).resolve()
    if can_delete and target_path.exists():
        shutil.rmtree(str(target_path))

    ctx.obj["app"].convert(target_path, name)


def invoke():
    cli(obj={})



from .utils import cmd


def test_run_check__counter():
    result = cmd("examples/counter.py", "run", "check")
    assert result.stdout.strip() == "System check identified no issues (0 silenced)."


def test_run_check__scale():
    result = cmd("examples/scale.py", "run", "check")
    assert result.stdout.strip() == "System check identified no issues (0 silenced)."



import urllib.request

from .utils import cmd, converted_process, runserver


TEST_APP = "scale"
TEST_SCRIPT = f"examples/{TEST_APP}.py"
TEST_BIND = "127.0.0.1:8042"
TEST_HTTP = f"http://{TEST_BIND}/"
TIMEOUT = 10


def test_runserver__fbv_with_model(tmp_path):
    cmd(TEST_SCRIPT, "run", "makemigrations", TEST_APP)
    cmd(TEST_SCRIPT, "run", "migrate")
    cmd(TEST_SCRIPT, "convert", str(tmp_path), "--name=converted", "--delete")

    with (
        converted_process(tmp_path, "runserver", TEST_BIND) as handle,
        runserver(handle),
    ):
        response = urllib.request.urlopen(TEST_HTTP, timeout=TIMEOUT)
        assert response.getcode() == 200
        assert "Number of page loads" in response.read().decode("utf-8")
